package io.xjar;

import java.io.FileInputStream;
import java.util.Set;
import java.util.Collections;
import java.util.HashSet;
import io.xjar.filter.XNotEntryFilter;
import io.xjar.filter.XAnyEntryFilter;
import java.util.Collection;
import io.xjar.filter.XAllEntryFilter;
import javax.crypto.SecretKey;
import io.xjar.key.XSymmetricSecureKey;
import java.security.SecureRandom;
import io.xjar.key.XSecureRandom;
import javax.crypto.KeyGenerator;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import io.xjar.key.XKey;
import java.io.OutputStreamWriter;
import java.io.FileOutputStream;
import java.io.File;
import java.io.Writer;
import java.io.Reader;
import java.io.Closeable;
import java.io.OutputStream;
import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;

public abstract class XKit implements XConstants
{
    public static byte[] readln(final InputStream in) throws IOException {
        int b = in.read();
        if (b == -1) {
            return null;
        }
        final ByteArrayOutputStream bos = new ByteArrayOutputStream();
        while (b != -1) {
            switch (b) {
                case 13: {
                    break;
                }
                case 10: {
                    return bos.toByteArray();
                }
                default: {
                    bos.write(b);
                    break;
                }
            }
            b = in.read();
        }
        return bos.toByteArray();
    }
    
    public static void writeln(final OutputStream out, final byte[] line) throws IOException {
        if (line == null) {
            return;
        }
        out.write(line);
        out.write(13);
        out.write(10);
    }
    
    public static byte[] read(final InputStream in) throws IOException {
        final ByteArrayOutputStream bos = new ByteArrayOutputStream();
        transfer(in, bos);
        return bos.toByteArray();
    }
    
    public static void write(final OutputStream out, final byte[] data) throws IOException {
        if (data == null) {
            return;
        }
        out.write(data);
    }
    
    public static void close(final Closeable closeable) {
        try {
            close(closeable, true);
        }
        catch (final IOException e) {
            throw new RuntimeException(e);
        }
    }
    
    public static void close(final Closeable closeable, final boolean quietly) throws IOException {
        if (closeable == null) {
            return;
        }
        try {
            closeable.close();
        }
        catch (final IOException e) {
            if (!quietly) {
                throw e;
            }
        }
    }
    
    public static long transfer(final InputStream in, final OutputStream out) throws IOException {
        long total = 0L;
        final byte[] buffer = new byte[4096];
        int length;
        while ((length = in.read(buffer)) != -1) {
            out.write(buffer, 0, length);
            total += length;
        }
        out.flush();
        return total;
    }
    
    public static long transfer(final Reader reader, final Writer writer) throws IOException {
        long total = 0L;
        final char[] buffer = new char[4096];
        int length;
        while ((length = reader.read(buffer)) != -1) {
            writer.write(buffer, 0, length);
            total += length;
        }
        writer.flush();
        return total;
    }
    
    public static long transfer(final InputStream in, final File file) throws IOException {
        OutputStream out = null;
        try {
            out = new FileOutputStream(file);
            return transfer(in, out);
        }
        finally {
            close(out);
        }
    }
    
    public static long transfer(final Reader reader, final File file) throws IOException {
        OutputStream out = null;
        Writer writer = null;
        try {
            out = new FileOutputStream(file);
            writer = new OutputStreamWriter(out);
            return transfer(reader, writer);
        }
        finally {
            close(writer);
            close(out);
        }
    }
    
    public static boolean delete(final File file) {
        return delete(file, false);
    }
    
    public static boolean delete(final File file, final boolean recursively) {
        if (file.isDirectory() && recursively) {
            boolean deleted = true;
            final File[] files = file.listFiles();
            for (int i = 0; files != null && i < files.length; ++i) {
                deleted &= delete(files[i], true);
            }
            return deleted && file.delete();
        }
        return file.delete();
    }
    
    public static XKey key(final String password) throws NoSuchAlgorithmException {
        return key("AES/CBC/PKCS5Padding", 128, 128, password);
    }
    
    public static XKey key(final String algorithm, final int keysize, final int ivsize, final String password) throws NoSuchAlgorithmException {
        final MessageDigest sha512 = MessageDigest.getInstance("SHA-512");
        final byte[] seed = sha512.digest(password.getBytes(StandardCharsets.UTF_8));
        final KeyGenerator generator = KeyGenerator.getInstance(algorithm.split("[/]")[0]);
        final XSecureRandom random = new XSecureRandom(seed);
        generator.init(keysize, (SecureRandom)random);
        final SecretKey key = generator.generateKey();
        generator.init(ivsize, (SecureRandom)random);
        final SecretKey iv = generator.generateKey();
        return (XKey)new XSymmetricSecureKey(algorithm, keysize, ivsize, password, key.getEncoded(), iv.getEncoded());
    }
    
    public static <E> XAllEntryFilter<E> all() {
        return (XAllEntryFilter<E>)new XAllEntryFilter();
    }
    
    public static <E> XAllEntryFilter<E> all(final Collection<? extends XEntryFilter<E>> filters) {
        return (XAllEntryFilter<E>)new XAllEntryFilter((Collection)filters);
    }
    
    public static <E> XAllEntryFilter<E> and() {
        return (XAllEntryFilter<E>)new XAllEntryFilter();
    }
    
    public static <E> XAllEntryFilter<E> and(final Collection<? extends XEntryFilter<E>> filters) {
        return (XAllEntryFilter<E>)new XAllEntryFilter((Collection)filters);
    }
    
    public static <E> XAnyEntryFilter<E> any() {
        return (XAnyEntryFilter<E>)new XAnyEntryFilter();
    }
    
    public static <E> XAnyEntryFilter<E> any(final Collection<? extends XEntryFilter<E>> filters) {
        return (XAnyEntryFilter<E>)new XAnyEntryFilter((Collection)filters);
    }
    
    public static <E> XAnyEntryFilter<E> or() {
        return (XAnyEntryFilter<E>)new XAnyEntryFilter();
    }
    
    public static <E> XAnyEntryFilter<E> or(final Collection<? extends XEntryFilter<E>> filters) {
        return (XAnyEntryFilter<E>)new XAnyEntryFilter((Collection)filters);
    }
    
    public static <E> XEntryFilter<E> not(final XEntryFilter<E> filter) {
        return (XEntryFilter<E>)new XNotEntryFilter((XEntryFilter)filter);
    }
    
    public static boolean isRelative(final String path) {
        return !isAbsolute(path);
    }
    
    public static boolean isAbsolute(final String path) {
        if (path.startsWith("/")) {
            return true;
        }
        final Set<File> roots = new HashSet<File>();
        Collections.addAll(roots, File.listRoots());
        File root;
        for (root = new File(path); root.getParentFile() != null; root = root.getParentFile()) {}
        return roots.contains(root);
    }
    
    public static String absolutize(final String path) {
        return normalize(isAbsolute(path) ? path : (System.getProperty("user.dir") + File.separator + path));
    }
    
    public static String normalize(final String path) {
        return path.replaceAll("[/\\\\]+", "/");
    }
    
    public static byte[] md5(final File file) throws IOException {
        try {
            return hash(file, MessageDigest.getInstance("MD5"));
        }
        catch (final NoSuchAlgorithmException e) {
            throw new IOException(e);
        }
    }
    
    public static byte[] sha1(final File file) throws IOException {
        try {
            return hash(file, MessageDigest.getInstance("SHA-1"));
        }
        catch (final NoSuchAlgorithmException e) {
            throw new IOException(e);
        }
    }
    
    public static byte[] hash(final File file, final MessageDigest hash) throws IOException {
        try (FileInputStream fis = new FileInputStream(file)) {
            final byte[] buf = new byte[8192];
            int len;
            while ((len = fis.read(buf)) != -1) {
                hash.update(buf, 0, len);
            }
            return hash.digest();
        }
    }
}
